package D2P::Parser::Multi;
use base 'D2P::Parser';

use	File::Basename;
use	D2P::Entity::Schema;


sub	init
{
	my	$this = shift;

	$this->SUPER::init();
	$this->{config}->{enable} = 0;
	$this->{config}->{defParser} = "StdParser";
}


###
## Create arguments list
###
sub	commandLine
{
	my	$this	= shift;

	return	(
		"-me|--multi-enable"	=> {
			desc	=> "Enable multiple files parser. Other Multi parser options will "
				. "be ignored if this one is not present.",
			bvar	=> "enable",
			uniq	=> 1
		},
		"-md|--multi-default-parser" => {
			desc	=> "Set default parser for multiple files parsing. Default is "
				. "StdParser.",
			args	=> ["defParser"],
			uniq	=> 1
		},
		"-mp|--multi-parser" => {
			desc	=> "Set parser to use for next input file. If not present the default "
				. "parser will be used.",
			argl	=> "parsers",
			args	=> ["parser"],
			uniq	=> 0
		},
		"-mi|--multi-input" => {
			desc	=> "Add an input file to the list of files to be parsed. If a -mp option"
				. " has been specified before the -mi option, the specified parsed will "
				. " be used, otherwise the default parser will be used.",
			argl	=> "infiles",
			args	=> ["infile"],
			uniq	=> 0,
			hdlr	=> "argInputFile"
		}
	);
}


###
## This method gets called whenever a "-mi" argument is found on the command line
## It ensures that the list of parsers gets updated when the list of files is.
###
sub	argInputFile
{
	my	$this	= shift;
	my	$argDsc = shift;

	push @{$this->{config}->{parsers}}, { parser => "", default => 1 }
		if (@{$this->{config}->{parsers}} < @{$this->{config}->{infiles}});
}


###
## Have other parser modules parse their input files, then merge their output
## schemas.
###
sub	parse
{
	my	$this	= shift;
	return 0 unless $this->{config}->{enable};

	## Get input files
	my	@fList	= map { $_->{infile} } @{$this->{config}->{infiles}};
	die "ERROR : No files configured for input ; use the -mi option\n"
		unless @fList;

	## Create parsers hash table and module list
	my	@pMList	= ();
	my	%pHash	= ();
	for my $i (0 .. (@fList - 1))
	{
		my	$fname	= $fList[$i];
		my	$bn	= File::Basename::basename($fname);
		$this->{file} .= ($i > 0 ? " ; " : "") . $bn;

		my	$parser	= $this->{config}->{parsers}->[$i];
		my	$pmod	= "";
		if	($parser->{default})
		{
			$pmod = $this->{config}->{defParser};
		}
		else
		{
			$pmod = $parser->{parser};
		}
		push(@pMList, $pmod) unless grep(/^$pmod$/, @pMList);

		$pHash{$fname} = {
			name	=> $bn,
			parser	=> $pmod,
			inst	=> undef
		};
	}

	## Try and load parser modules
	foreach my $mod (@pMList)
	{
		die "ERROR : multiple files parser may not use itself as a parser\n"
			if $mod eq "Multi";

		my	$cmd = 'use D2P::Parser::' . $mod . ';';
		eval $cmd;
		die "ERROR : could not find parser module $mod\n" if $@;
	}

	## Initialize a parser for each input file, then parse the file
	my	$parseFailed = 0;
	foreach my $fname (keys %pHash)
	{
		my	$pname	= "D2P::Parser::" . $pHash{$fname}->{parser};
		my	$inst	= $pHash{$fname}->{inst} = new $pname();
		die "ERROR : could not initialize parser for $fname (using module $pname)\n"
			unless ref($pHash{$fname}->{inst}) eq $pname;

		my	$pRes = 0;
		$inst->file($fname);
		eval '$pRes = $inst->parse();';
		if	($@)
		{
			my	$pError = $@;
			print STDERR "An error occured when parsing file $fname\n";
			print STDERR "$pname instance returned the following :\n";
			print STDERR "$pError\n";
			print STDERR "\n" unless $pError =~ /\n$/;
			$parseFailed = 1;
		}
		elsif	(!$pRes || !$inst->{schema})
		{
			print STDERR "An error occured when parsing file $fname\n";
			print STDERR "$pname instance failed to parse it, but did not\n";
			print STDERR "return any error.\n\n";
			$parseFailed = 1;
		}
	}
	die "Parser stopped because of errors\n" if $parseFailed;

	## Create a list of all schemas
	my	@schemas = map { $pHash{$_}->{inst}->schema() } keys %pHash;

	## Create local schema, then merge all generated schemas into the new one
	$this->{schema} = new D2P::Entity::Schema();
	return	$this->mergeSchemas($this->{schema}, @schemas);
}


###
## Merge all of the generated schemas into a single schema
###
sub	mergeSchemas
{
	my	$this		= shift;
	my	$dest		= shift;
	my	@sources	= @_;
	return 0 unless @sources;

	## Create hash tables containing all modules and external modules
	# from the sources
	my	$status		= 1;
	my	%modules	= ();
	my	%extMods	= ();
	my	%extModsDel	= ();
	foreach my $schema (@sources)
	{
		$dest->header($schema->header()) unless $dest->header();
		foreach my $mod (@{$schema->modules()->{ordered}})
		{
			my	$mn = $mod->{name};
			if	($modules{$mn})
			{
				print STDERR "ERROR : module '$mn' defined more than once\n";
				$status = 0;
				next;
			}
			$modules{$mn} = $mod;
		}

		foreach my $emn (keys %{$schema->extModules()->{byname}})
		{
			my	$emod = $schema->extModules()->{byname}->{$emn};
			$extMods{$emn} = [] unless $extMods{$emn};
			push @{$extMods{$emn}}, $emod;
		}
	}

	## Check for external modules that exist in the modules list
	foreach my $em (keys %extMods)
	{
		next unless $modules{$em};
		$extModsDel{$em} = $extMods{$em};
		undef $extMods{$em};
	}

	## Add modules to current schema
	foreach my $mn (keys %modules)
	{
		$dest->addModuleCopy($modules{$mn});
	}

	## Merge external modules
	foreach $mn (keys %extMods)
	{
		my	$emods	= $extMods{$mn};
		next unless @$emods;

		my	$newmod	= 0;
		my	$tables = { };
		foreach my $emod (@$emods)
		{
			my	$th = $emod->tables()->{byname};
			foreach my $tname (keys %$th)
			{
				next if $tables{$tname};
				my $t = $tables{$tname} = $th->{$tname};
				$newmod->addTable($t) if $newmod;
			}
			$newmod = $emod unless $newmod;
		}
		$dest->addModuleCopy($newmod);
	}

	## Modify relations
	foreach my $rel (values %{$dest->{relations}})
	{
		if	(!$dest->{tables}->{byid}->{$rel->fromId()})
		{
			my	$tn = $rel->from()->name();
			my	$t = $dest->{tables}->{byname}->{$tn};
			$rel->{fromId} = $t->diaId() if defined $t;
		}
		if	(!$dest->{tables}->{byid}->{$rel->toId()})
		{
			my	$tn = $rel->to()->name();
			my	$t = $dest->{tables}->{byname}->{$tn};
			$rel->{toId} = $t->diaId() if defined $t;
		}
	}

	$dest->postprocess();
	return	$status;
}


1;
